@@ -19,13 +19,14 @@ from __future__ import annotations
 
 import typing
 from typing import Any
-from typing import Optional
-from typing import Set
 from typing import NamedTuple
+from typing import Optional
 from typing import Union
 
-import time
+import logging
 import random
+import time
+from collections.abc import Sequence
 from dataclasses import dataclass
 from dataclasses import field
 from datetime import datetime
@@ -33,23 +34,31 @@ from datetime import datetime
 from gi.repository import Gio
 from gi.repository import GLib
 
-from nbxmpp.simplexml import Node
-from nbxmpp.namespaces import Namespace
-from nbxmpp.protocol import Protocol
-from nbxmpp.protocol import JID
+from nbxmpp import exceptions
 from nbxmpp.const import AdHocAction
-from nbxmpp.const import IqType
 from nbxmpp.const import AdHocNoteType
-from nbxmpp.const import MessageType
-from nbxmpp.const import AvatarState
-from nbxmpp.const import StatusCode
-from nbxmpp.const import PresenceType
 from nbxmpp.const import AdHocStatus
-from nbxmpp.const import InviteType
-from nbxmpp.const import Role
 from nbxmpp.const import Affiliation
 from nbxmpp.const import AnonymityMode
+from nbxmpp.const import AvatarState
+from nbxmpp.const import InviteType
+from nbxmpp.const import IqType
+from nbxmpp.const import MessageType
 from nbxmpp.const import PresenceShow
+from nbxmpp.const import PresenceType
+from nbxmpp.const import Role
+from nbxmpp.const import StatusCode
+from nbxmpp.language import LanguageMap
+from nbxmpp.language import LanguageRange
+from nbxmpp.language import LanguageTag
+from nbxmpp.modules.fallback import FallbacksForT
+from nbxmpp.modules.fallback import strip_fallback
+from nbxmpp.namespaces import Namespace
+from nbxmpp.protocol import JID
+from nbxmpp.protocol import Protocol
+from nbxmpp.simplexml import Node
+
+log = logging.getLogger('nbxmpp.structs')
 
 
 class StanzaHandler(NamedTuple):
@@ -175,8 +184,8 @@ class MoodData(NamedTuple):
 
 
 class BlockingPush(NamedTuple):
-    block: Set[JID]
-    unblock: Set[JID]
+    block: set[JID]
+    unblock: set[JID]
     unblock_all: bool
 
 
@@ -276,9 +285,9 @@ class AdHocCommandNote(NamedTuple):
 class IBBData(NamedTuple):
     type: str
     sid: str
-    block_size: Optional[int]
-    seq: Optional[int]
-    data: Optional[str]
+    block_size: Optional[int] = None
+    seq: Optional[int] = None
+    data: Optional[str] = None
 
 
 class OOBData(NamedTuple):
@@ -291,23 +300,25 @@ class CorrectionData(NamedTuple):
 
 
 class ReplyData(NamedTuple):
-    to: str
+    to: JID
     id: str
-    fallback_start: Optional[int]
-    fallback_end: Optional[int]
 
 
 class ModerationData(NamedTuple):
     stanza_id: str
-    moderator_jid: Optional[str]
     by: Optional[JID]
     reason: Optional[str]
-    timestamp: Optional[str]
-    stamp: float
+    stamp: datetime
     is_tombstone: bool
     occupant_id: Optional[str]
 
 
+class RetractionData(NamedTuple):
+    id: Optional[str]
+    is_tombstone: bool
+    timestamp: Optional[float]
+
+
 class DiscoItems(NamedTuple):
     jid: JID
     node: str
@@ -379,7 +390,7 @@ class RosterItem:
     ask: Optional[str] = None
     subscription: Optional[str] = None
     approved: Optional[str] = None
-    groups: Set[str] = field(default_factory=set)
+    groups: set[str] = field(default_factory=set)
 
     @classmethod
     def from_node(cls, node: Node) -> RosterItem:
@@ -502,7 +513,18 @@ class DiscoInfo(NamedTuple):
 
     @property
     def has_message_moderation(self) -> bool:
-        return Namespace.MESSAGE_MODERATE in self.features
+        return (
+            Namespace.MESSAGE_MODERATE in self.features or
+            Namespace.MESSAGE_MODERATE_1 in self.features
+        )
+
+    @property
+    def moderation_namespace(self) -> Optional[str]:
+        if Namespace.MESSAGE_MODERATE_1 in self.features:
+            return Namespace.MESSAGE_MODERATE_1
+        if Namespace.MESSAGE_MODERATE in self.features:
+            return Namespace.MESSAGE_MODERATE
+        return None
 
     @property
     def is_muc(self) -> bool:
@@ -625,10 +647,7 @@ class DiscoInfo(NamedTuple):
 
     @property
     def is_gateway(self) -> bool:
-        for identity in self.identities:
-            if identity.category == 'gateway':
-                return True
-        return False
+        return any(identity.category == 'gateway' for identity in self.identities)
 
     @property
     def gateway_name(self) -> Optional[str]:
@@ -645,10 +664,7 @@ class DiscoInfo(NamedTuple):
         return None
 
     def has_category(self, category: str) -> bool:
-        for identity in self.identities:
-            if identity.category == category:
-                return True
-        return False
+        return any(identity.category == category for identity in self.identities)
 
     def has_identity(self, category: str, type_: str) -> bool:
         for identity in self.identities:
@@ -707,7 +723,7 @@ class AdHocCommand(NamedTuple):
     sessionid: Optional[str] = None
     status: Optional[AdHocStatus] = None
     data: Optional[Node] = None
-    actions: Optional[Set[AdHocAction]] = None
+    actions: Optional[set[AdHocAction]] = None
     default: Optional[AdHocAction] = None
     notes: Optional[list[AdHocCommandNote]] = None
 
@@ -974,6 +990,43 @@ class ReceiptData(NamedTuple):
         return self.type == 'received'
 
 
+@dataclass
+class Hat:
+    uri: str
+    title: str
+
+
+class HatData:
+    def __init__(self) -> None:
+        self._hat_map = LanguageMap()
+
+    def add_hat(self, hat: Hat, lang: str | None) -> None:
+        language_tag = LanguageTag(tag=lang) if lang else None
+        if language_tag not in self._hat_map:
+            self._hat_map[language_tag] = []
+        self._hat_map[language_tag].append(hat)
+
+    def get_hats(
+        self,
+        language_range: Sequence[LanguageRange] | None = None
+    ) -> list[Hat]:
+
+        if language_range is None:
+            return self._hat_map.any()
+
+        try:
+            return self._hat_map.lookup(language_range)
+        except KeyError:
+            return self._hat_map.any()
+
+
+@dataclass
+class EncryptionData:
+    protocol: str
+    key: str | None
+    trust: int
+
+
 class Properties:
     pass
 
@@ -992,6 +1045,7 @@ class MessageProperties:
     remote_jid: Optional[JID] = None
     subject = None
     body: Optional[str] = None
+    bodies: Optional[BodyData] = None
     thread: Optional[str] = None
     user_timestamp = None
     timestamp: float = field(default_factory=time.time)
@@ -1004,7 +1058,7 @@ class MessageProperties:
     occupant_id: Optional[str] = None
     muc_jid: Optional[JID] = None
     muc_nickname: Optional[str] = None
-    muc_status_codes: Optional[Set[StatusCode]] = None
+    muc_status_codes: Optional[set[StatusCode]] = None
     muc_private_message: bool = False
     muc_invite = None
     muc_decline = None
@@ -1019,7 +1073,7 @@ class MessageProperties:
     pubsub_event: Optional[PubSubEventData] = None
     openpgp = None
     omemo = None
-    encrypted = None
+    encrypted: EncryptionData | None = None
     pgp_legacy = None
     marker: Optional[ChatMarker] = None
     receipt: Optional[ReceiptData] = None
@@ -1027,6 +1081,7 @@ class MessageProperties:
     correction: Optional[CorrectionData] = None
     reply_data: Optional[ReplyData] = None
     moderation: Optional[ModerationData] = None
+    retraction: Optional[RetractionData] = None
     attention: bool = False
     forms = None
     xhtml: Optional[str] = None
@@ -1141,6 +1196,10 @@ class MessageProperties:
     def is_moderation(self) -> bool:
         return self.moderation is not None
 
+    @property
+    def is_retraction(self) -> bool:
+        return self.retraction is not None
+
     @property
     def has_attention(self) -> bool:
         return self.attention
@@ -1234,11 +1293,12 @@ class PresenceProperties:
     avatar_sha: Optional[str] = None
     avatar_state: AvatarState = AvatarState.IGNORE
     muc_jid: Optional[JID] = None
-    muc_status_codes: Optional[Set[StatusCode]] = None
+    muc_status_codes: Optional[set[StatusCode]] = None
     muc_user: Optional[MucUserData] = None
     muc_nickname: Optional[str] = None
     muc_destroyed: Optional[MucDestroyed] = None
     entity_caps: Optional[EntityCapsData] = None
+    hats: Optional[HatData] = None
 
     @property
     def is_self_presence(self) -> bool:
@@ -1346,3 +1406,62 @@ class XHTMLData:
         if body is not None:
             return str(body)
         return str(self._bodys.popitem()[1])
+
+
+@dataclass
+class ChannelBindingData:
+    type: str
+    data: bytes
+
+
+@dataclass
+class MDSData:
+    jid: JID
+    stanza_id: str
+    stanza_id_by: JID
+
+
+@dataclass
+class BodyData:
+    def __init__(
+        self,
+        stanza: Node,
+        fallbacks_for: FallbacksForT | None = None,
+        fallback_ns: set[str] | None = None,
+    ) -> None:
+
+        self._body_map = LanguageMap()
+        self._fallbacks_for = fallbacks_for
+        self._fallback_ns = fallback_ns
+
+        for body in stanza.getTags('body'):
+            lang = body.getAttr('xml:lang')
+            lang_tag = LanguageTag(tag=lang) if lang else None
+            self._body_map[lang_tag] = (lang, body.getData())
+
+    def get(
+        self,
+        language_range: Sequence[LanguageRange] | None,
+    ) -> str:
+
+        if not self._body_map:
+            return ''
+
+        if language_range is None:
+            lang, text = self._body_map.any()
+
+        else:
+            try:
+                lang, text = self._body_map.lookup(language_range)
+            except KeyError:
+                lang, text = self._body_map.any()
+
+        if not self._fallbacks_for or not self._fallback_ns:
+            return text
+
+        try:
+            return strip_fallback(
+                self._fallbacks_for, self._fallback_ns, lang, text)
+        except exceptions.FallbackLanguageError:
+            log.warning('Missing fallback for language: %s', lang)
+            return text
